# -*- coding: utf-8 -*-
#
#       infoview.py
#       
#       Copyright © 2011 Brandon Invergo <b.invergo@gmail.com>
#       
#       This file is part of Grotesque.
#
#       Grotesque is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       Grotesque is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with Grotesque.  If not, see <http://www.gnu.org/licenses/>.


import os.path

from gi.repository import Gtk, Pango


class InfoView(Gtk.TextView):
    '''This class implements a widget which displays story information.
    
    '''
    def __init__(self):
        # Create all of the tags which will markup the text:
        # Title: XX-large & bold
        title_tag = Gtk.TextTag(name='title')
        title_tag.set_property('scale', 1.2 ** 3)
        title_tag.set_property('weight', Pango.Weight.BOLD)
        # Headline: X-Large & italic
        headline_tag = Gtk.TextTag(name='headline')
        headline_tag.set_property('scale', 1.2 ** 2)
        headline_tag.set_property('style', Pango.Style.ITALIC)
        # Series: X-Large
        series_tag = Gtk.TextTag(name='series')
        series_tag.set_property('scale', 1.2 ** 2)
        # Author: Large
        author_tag = Gtk.TextTag(name='author')
        author_tag.set_property('scale', 1.2)
        # Info: Small & italic
        info_tag = Gtk.TextTag(name='info')
        info_tag.set_property('scale', 0.83)
        info_tag.set_property('style', Pango.Style.ITALIC)
        text_buffer = Gtk.TextBuffer()
        tag_table = text_buffer.get_tag_table()
        for tag in [title_tag, headline_tag, series_tag, author_tag, info_tag]:
            tag_table.add(tag)
        super(InfoView, self).__init__(buffer=text_buffer)

    def render_story(self, story):
        '''This method inserts the story information in the text buffer and
        applies the markup tags to properly format it.
        
        '''
        self.clear()
        text_buffer = self.get_buffer()
        text_iter = text_buffer.get_start_iter()
        tag_table = text_buffer.get_tag_table()
        text_buffer.insert_with_tags(
            text_iter, '{0}\n'.encode('UTF-8').format(story.title),
            tag_table.lookup('title'))
        if story.headline:
            text_buffer.insert_with_tags(
                text_iter, '{0}\n'.encode('UTF-8').format(story.headline),
                tag_table.lookup('headline'))
        if story.series != 'None':
            text_buffer.insert_with_tags(
                text_iter, '{0} (part {1})\n'.encode('UTF-8').format(
                story.series, story.seriesnumber), tag_table.lookup('series'))
        text_buffer.insert_with_tags(
            text_iter, 'by {0}\n'.encode('UTF-8').format(story.author),
            tag_table.lookup('author'))
        if story.group != 'None':
            text_buffer.insert(
                text_iter, '({0})\n'.encode('UTF-8').format(story.group))
        rating_pixbuf = self.get_rating_pixbuf(story.starrating)
        text_buffer.insert_pixbuf(text_iter, rating_pixbuf)
        text_buffer.insert(text_iter, '\n')
        if story.description:
            text_buffer.insert(
                text_iter, '\n{0}\n'.encode('UTF-8').format(story.description))
        text_buffer.insert_with_tags(
            text_iter, '\nFirst published: {0}\n'.encode('UTF-8').format(
            story.firstpublished), tag_table.lookup('info'))
        text_buffer.insert_with_tags(
            text_iter, 'Genre: {0}\n'.encode('UTF-8').format(story.genre),
            tag_table.lookup('info'))
        text_buffer.insert_with_tags(
            text_iter, 'Forgiveness: {0}\n'.encode('UTF-8').format(
            story.forgiveness), tag_table.lookup('info'))
        text_buffer.insert_with_tags(
            text_iter, 'Format: {0}\n'.encode('UTF-8').format(story.format),
            tag_table.lookup('info'))
        text_buffer.insert_with_tags(
            text_iter, 'IFID: {0}'.encode('UTF-8').format(story.ifid),
            tag_table.lookup('info'))

    def clear(self):
        '''This is a helper method to clear the contents of the view.
        
        '''
        text_buffer = self.get_buffer()
        text_buffer.set_text('')

    def get_rating_pixbuf(self, rating):
        '''This method fetches the star rating image given a textual
        description of it ("2", "3.5", etc.).
        
        '''
        rating_str = rating.replace('.', '-')
        rating_file = os.path.join(os.path.dirname(__file__), 'data',
                                   'stars_{0}.png'.format(rating_str))
        rating_image = Gtk.Image.new_from_file(rating_file)
        return rating_image.get_pixbuf()
