# -*- coding: utf-8 -*-
#
#       library.py
#       
#       Copyright © 2009, 2010 Per Liedman <per@liedman.net>
#       Copyright © 2011 Brandon Invergo <b.invergo@gmail.com>
#       
#       This file is part of Grotesque.
#
#       Grotesque is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       Grotesque is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with Grotesque.  If not, see <http://www.gnu.org/licenses/>.


import cPickle
import os.path
from xml.dom import getDOMImplementation

from gi.repository import Gtk, Gdk, GdkPixbuf, Pango

from filterstore import FilterStore
from treatyofbabel import ifiction, storydata


class Library:
    '''This class stores the user's library of interactive fiction.
    
    '''
    def __init__(self):
        # The main list store contains the following columns:
        # 0: played state, 1: title, 2: author, 3: language, 4: headline,
        # 5: publishing date, 6: genre, 7: group, 8: series, 9: series number,
        # 10: forgiveness, 11: star rating image, 12: star rating text,
        # 13: description, 14: story file location, 15: story object,
        # 16: the text weight to render the row at
        self.list_store = Gtk.ListStore(bool, str, str, str, str, str, str,
                                        str, str, str, str, GdkPixbuf.Pixbuf,
                                        str, str, str, object, Pango.Weight)
        # The following Filterstores keep track of which authors, years, etc
        # are currently represented in the library so the user may filter it by
        # them.
        self.author_store = FilterStore()
        self.author_store.append(['(All)'])
        self.year_store = FilterStore()
        self.year_store.append(['(All)'])
        self.genre_store = FilterStore()
        self.genre_store.append(['(All)'])
        self.group_store = FilterStore()
        self.group_store.append(['(All)'])
        self.series_store = FilterStore()
        self.series_store.append(['(All)'])
        self.forgiveness_store = FilterStore()
        self.forgiveness_store.append(['(All)'])
        self.rating_store = FilterStore(True)
        # This list keeps track of which files are already in the library.
        self._files = []

    def add_entry(self, entry):
        '''This method handles adding stories to the library.
        
        '''
        # Avoid duplicate entries.
        if entry.story_file in self._files:
            return
        # For some of the fields, it's better to have some default text rather
        # than no text.
        if not entry.author:
            entry.author = 'Unknown'
        if not entry.firstpublished:
            entry.firstpublished = 'Unknown'
        if not entry.genre:
            entry.genre = 'Unknown'
        if not entry.group:
            entry.group = 'None'
        if not entry.series:
            entry.series = 'None'
        if not entry.forgiveness:
            entry.forgiveness = 'Unknown'
        if not entry.starrating:
            entry.starrating = '0'
        #Gdk.threads_enter()
        # Set the font weight depending on whether the story has been played or
        # not
        if entry.played == False:
            text_weight = Pango.Weight.BOLD
        else:
            text_weight = Pango.Weight.NORMAL
        # Get the star rating image
        rating_image = self.get_rating_pixbuf(entry.starrating)
        # Add the story to the main list.
        self.list_store.append([entry.played, entry.title, entry.author,
                                entry.language, entry.headline,
                                entry.firstpublished, entry.genre, entry.group,
                                entry.series, entry.seriesnumber,
                                entry.forgiveness, rating_image,
                                entry.starrating, entry.description,
                                entry.story_file, entry, text_weight])
        # Add it to the list of known files.
        self._files.append(entry.story_file)
        # Parse the authors text to get a list of all the authors involved.
        authors = self.parse_list_str(entry.author)
        for author in authors:
            # Only add the author's real name to the filter (no pen names).
            author_real_name = author.split('(')[0]
            if author_real_name == '':
                continue
            self.author_store.add_item(author_real_name.strip())
        # We only filter by year, so if month/day are included, cut them out.
        year = entry.firstpublished.split('-')[0]
        self.year_store.add_item(year)
        # Parse genre to get a list of all genres named.
        genres = self.parse_list_str(entry.genre)
        for genre in genres:
            self.genre_store.add_item(genre.strip())
        self.group_store.add_item(entry.group)
        self.series_store.add_item(entry.series)
        self.forgiveness_store.add_item(entry.forgiveness)
        self.rating_store.add_item(entry.starrating, rating_image)
        #Gdk.threads_leave()

    def edit_entry(self, row_iter, biblio_info):
        '''This method handles the editing of a story entry.
        
        '''
        entry = self.get_story(row_iter)
        if entry.title != biblio_info['Title']:
            if not biblio_info['Title']:
                entry.title = 'Unknown'
            else:
                entry.title = biblio_info['Title']
            self.list_store.set_value(row_iter, 1, entry.title)
        if entry.author != biblio_info['Author']:
            if entry.author:
                authors = self.parse_list_str(entry.author)
                for author in authors:
                    self.author_store.remove_item(entry.author.lower())
            if not biblio_info['Author']:
                entry.author = 'Unknown'
            else:
                entry.author = biblio_info['Author']
            self.list_store.set_value(row_iter, 2, entry.author)
            authors = self.parse_list_str(entry.author)
            for author in authors:
                self.author_store.add_item(entry.author.lower())
        if entry.language != biblio_info['Language']:
            if not biblio_info['Language']:
                entry.language = None
            else:
                entry.language = biblio_info['Language']
            self.list_store.set_value(row_iter, 3, entry.language)
        if entry.headline != biblio_info['Headline']:
            if not biblio_info['Headline']:
                entry.headline = None
            else:
                entry.headline = biblio_info['Headline']
            self.list_store.set_value(row_iter, 4, entry.headline)
        if entry.firstpublished != biblio_info['First published']:
            if entry.firstpublished:
                year = entry.firstpublished.split('-')[0]
                self.year_store.remove_item(year)
            if not biblio_info['First published']:
                entry.firstpublished = 'Unknown'
            else:
                entry.firstpublished = biblio_info['First published']
            self.list_store.set_value(row_iter, 5, entry.firstpublished)
            year = entry.firstpublished.split('-')[0]
            self.year_store.add_item(year)
        if entry.genre != biblio_info['Genre']:
            if entry.genre:
                genres = self.parse_list_str(entry.genre)
                for genre in genres:
                    self.genre_store.remove_item(genre.lower())
            if not biblio_info['Genre']:
                entry.genre = 'Unknown'
            else:
                entry.genre = biblio_info['Genre']
            self.list_store.set_value(row_iter, 6, entry.genre)
            genres = self.parse_list_str(entry.genre)
            for genre in genres:
                self.genre_store.add_item(genre.lower())
        if entry.group != biblio_info['Group']:
            if entry.group:
                self.group_store.remove_item(entry.group.lower())
            if not biblio_info['Group']:
                entry.group = 'None'
            else:
                entry.group = biblio_info['Group']
            self.list_store.set_value(row_iter, 7, entry.group)
            self.group_store.add_item(entry.group.lower())
        if entry.series != biblio_info['Series']:
            if entry.series:
                self.series_store.remove_item(entry.series.lower())
            if not biblio_info['Series']:
                entry.series = 'None'
            else:
                entry.series = biblio_info['Series']
            self.list_store.set_value(row_iter, 8, entry.series)
            self.series_store.add_item(entry.series.lower())
        if entry.seriesnumber != biblio_info['Series number']:
            if not biblio_info['Series number']:
                entry.seriesnumber = None
            else:
                entry.seriesnumber = biblio_info['Series number']
            self.list_store.set_value(row_iter, 9, entry.seriesnumber)
        if entry.forgiveness != biblio_info['Forgiveness']:
            if entry.forgiveness:
                self.forgiveness_store.remove_item(entry.forgiveness.lower())
            if not biblio_info['Forgiveness']:
                entry.forgiveness = 'Unknown'
            else:
                entry.forgiveness = biblio_info['Forgiveness']
            self.list_store.set_value(row_iter, 10, entry.forgiveness)
            self.forgiveness_store.add_item(entry.forgiveness.lower())
        if entry.starrating != biblio_info['Rating']:
            if entry.starrating:
                self.rating_store.remove_item(entry.starrating)
            if not biblio_info['Rating']:
                entry.starrating = "0"
            else:
                entry.starrating = biblio_info['Rating']
            self.list_store.set_value(row_iter, 12, entry.starrating)
            rating_pixbuf = self.get_rating_pixbuf(entry.starrating)
            self.list_store.set_value(row_iter, 11, rating_pixbuf)
            self.rating_store.add_item(entry.starrating, rating_pixbuf)
        if entry.description != biblio_info['Description']:
            entry.description = biblio_info['Description']
            self.list_store.set_value(row_iter, 13, entry.description)
        if entry.format != biblio_info['Format']:
            entry.format = biblio_info['Format']
        if entry.ifid != biblio_info['IFID']:
            entry.ifid = biblio_info['IFID']
        self.list_store.set_value(row_iter, 15, entry)

    def get_rating_pixbuf(self, rating):
        '''This method fetches a star rating image given a text description
        of the rating (ie "1", "2.5", etc).
        
        '''
        rating_str = rating.replace('.', '-')
        rating_file = os.path.join(os.path.dirname(__file__), 'data',
                                   'stars_{0}.png'.format(rating_str))
        rating_image = Gtk.Image.new_from_file(rating_file)
        return rating_image.get_pixbuf()

    def parse_list_str(self, list_str):
        '''This method parses strings which may be lists separated by commas
        or slashes, such as author lists or genre lists.
        
        '''
        itemized = list_str.split(',')
        final_list = itemized
        for item in itemized:
            for linker in [' and ', '&', '/']:
                if linker in item:
                    item_split = item.split(linker)
                    final_list.remove(item)
                    final_list.extend(item_split)
        return final_list

    def mark_story_played(self, row_iter, played=True):
        '''This method handles marking a story as having been played.
        
        '''
        self.list_store.set_value(row_iter, 0, played)
        story = self.list_store.get_value(row_iter, 15)
        story.played = played
        self.list_store.set_value(row_iter, 15, story)
        if played:
            weight = Pango.Weight.NORMAL
        else:
            weight = Pango.Weight.BOLD
        self.list_store.set_value(row_iter, 16, weight)

    def get_story(self, row_iter):
        '''This is a small helper method to return a story file.
        
        '''
        return self.list_store.get_value(row_iter, 15)

    def story_in_library(self, story_file):
        '''This is a small helper method to check if a story file is already in
        the library.
        
        '''
        return story_file in self._files

    def load_ifiction(self, file_handle):
        '''This method loads a library from an ifiction XML file.
        
        '''
        xml = file_handle.read()
        stories = ifiction.get_stories_from_ifiction(xml)
        for story in stories:
            self.add_entry(story)

    def save_to_ifiction(self, file_handle):
        '''This method saves the library to an ifiction XML file.
        
        '''
        impl = getDOMImplementation()
        doc = impl.createDocument(None, 'ifindex', None)
        root = doc.documentElement
        root.setAttribute('version', '1.0')
        root.setAttribute('xmlns',
                          'http://babel.ifarchive.org/protocol/iFiction/')
        list_iter = self.list_store.get_iter_first()
        while list_iter != None:
            story = self.list_store.get_value(list_iter, 15)
            root.appendChild(story.to_ifiction(doc))
            list_iter = self.list_store.iter_next(list_iter)
        xml = doc.toxml()
        file_handle.write(xml.encode('UTF-8'))

    def serialize_liststore(self):
        '''This method creates a serializable version of the list store.
        
        '''
        rows = []
        row_iter = self.list_store.get_iter_first()
        while row_iter:
            row = self.list_store.get(row_iter, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 
                                      10, 11, 12, 13, 14, 15, 16)
            row = list(row)
            row[11] = None
            story = row[15]
            row[15] = story.serialize()
            if row[16] == Pango.Weight.BOLD:
                row[16] = 'bold'
            else:
                row[16] = 'normal'
            rows.append(row)
            row_iter = self.list_store.iter_next(row_iter)
        return rows

    def deserialize_liststore(self, data):
        '''This method takes the serialized data and puts it back into the list
        store.
        
        '''
        for row in data:
            rating = row[12]
            pixbuf = self.get_rating_pixbuf(rating)
            row[11] = pixbuf
            story = row[15]
            row[15] = storydata.StoryData(story)
            if row[16] == 'bold':
                row[16] = Pango.Weight.BOLD
            else:
                row[16] = Pango.Weight.NORMAL
            self.list_store.append(row)
            self._files.append(row[14])

    def save(self, file_handle, version):
        '''This method saves the library to a file via pickling serialized
        versions of its data.
        
        '''
        data = {'filters': {}, 'version': version}
        data['filters']['authors'] = self.author_store.serialize()
        data['filters']['years'] = self.year_store.serialize()
        data['filters']['genres'] = self.genre_store.serialize()
        data['filters']['groups'] = self.group_store.serialize()
        data['filters']['series'] = self.series_store.serialize()
        data['filters']['forgiveness'] = self.forgiveness_store.serialize()
        data['filters']['ratings'] = self.rating_store.serialize()
        data['library'] = self.serialize_liststore()
        cPickle.dump(data, file_handle, cPickle.HIGHEST_PROTOCOL)

    def load(self, file_handle):
        '''Deprecated
        
        '''
        stories = cPickle.load(file_handle)
        for story in stories:
            self.add_entry(story)
