# -*- coding: utf-8 -*-
#
#       main.py
#       
#       Copyright © 2009, 2010 Per Liedman <per@liedman.net>
#       Copyright © 2011 Brandon Invergo <b.invergo@gmail.com>
#       
#       This file is part of Grotesque.
#
#       Grotesque is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       Grotesque is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with Grotesque.  If not, see <http://www.gnu.org/licenses/>.


import os.path
import cPickle
import errno
import shutil
import time
from subprocess import Popen

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GdkPixbuf, GObject

from library import Library
from settings import Settings
from settingsdialog import SettingsDialog
from infopaned import InfoPaned
from librarypaned import LibraryPaned
from storyimportthread import StoryImportThread
import libraryloadthread
import storyremovethread
from progressdialog import ProgressDialog
from treatyofbabel import ifdb


GROTESQUE_VERSION = "0.9.2"


class MainWindow(Gtk.Window):
    '''This class implements the main Grotesque window.
    
    '''
    def __init__(self, library_filename, settings):
        super(MainWindow, self).__init__(type=Gtk.WindowType.TOPLEVEL)
        # init_complete is used to block certain callbacks from happening while
        # the window is still being constructed.
        self.init_complete = False
        self.set_title('Grotesque')
        self.set_icon_from_file(os.path.join(os.path.dirname(__file__), 'data',
                                             'grotesque_icon.png'))
        self.connect('delete_event', self.on_close)
        self.settings = settings
        dimensions = settings.get_window_size()
        self.set_default_size(dimensions[0], dimensions[1])
        self.connect('key_press_event', self.on_key_pressed)

        # Create the library and its display widgets.
        self.library = Library()
        if not os.path.exists(library_filename):
            basename = os.path.splitext(library_filename)
            xml_library = '.'.join([basename[0], 'ifiction'])
            if os.path.exists(xml_library):
            # The user has upgraded from an old version of Grotesque which used
            # the XML ifiction format to store the library.
                coverart_dir = self.settings.coverart_dir
                import updatexmllibrary
                updatexmllibrary.update_library(self.library, xml_library,
                                                coverart_dir)
        else:
            self.load_library(library_filename)
        self.library_paned = LibraryPaned(self.library, self.settings)
        self.library_paned.connect('selection-changed',
                                   self.on_library_selection_changed)
        self.library_paned.connect('double-clicked',
                                   self.on_library_doubleclick)
        filter_percentage = self.settings.get_filter_percentage()
        self.library_paned.set_position(int(dimensions[0] * filter_percentage))

        # Create the info (story description & coverart) area 
        vpaned_percentage = self.settings.get_vpaned_percentage()
        self.info_paned = InfoPaned(self.settings)
        self.info_paned.connect('notify', self.on_infopaned_notify)
        self.info_paned.set_position(int(dimensions[1] - (dimensions[1] *
                                                          vpaned_percentage)))
        self.info_paned_block = False

        # Pack it all together.
        top_frame = Gtk.Frame()
        self.bottom_frame = Gtk.Frame()
        top_frame.set_shadow_type(Gtk.ShadowType.IN)
        self.bottom_frame.set_shadow_type(Gtk.ShadowType.IN)
        top_frame.add(self.library_paned)
        self.bottom_frame.add(self.info_paned)
        self.info_paned.show_coverart(self.settings.get_disp_coverart())
        toolbar = self.create_toolbar()
        top_frame_box = Gtk.VBox()
        top_frame_box.pack_start(toolbar, False, False, 8)
        top_frame_box.pack_start(top_frame, True, True, 8)

        self.vpaned = Gtk.VPaned()
        self.vpaned.connect('notify', self.on_vpaned_notify)
        self.vpaned.pack1(top_frame_box, True, False)
        self.vpaned.pack2(self.bottom_frame, False, False)
        self.vpaned.set_position(int(dimensions[1] * vpaned_percentage))
        self.vpaned_block = False
        self.add(self.vpaned)

        self.create_edit_dialog()

        self.show_all()
        self.init_complete = True

    def create_toolbar(self):
        '''This method creates the main toolbar.

        '''
        toolbar = Gtk.Toolbar()
        toolbar.get_style_context().add_class(Gtk.STYLE_CLASS_PRIMARY_TOOLBAR)
        self.button_play = Gtk.ToolButton(stock_id=Gtk.STOCK_MEDIA_PLAY)
        separator1 = Gtk.SeparatorToolItem()
        separator2 = Gtk.SeparatorToolItem()
        button_add = Gtk.ToolButton(stock_id=Gtk.STOCK_ADD)
        button_dir = Gtk.ToolButton(stock_id=Gtk.STOCK_DIRECTORY)
        self.button_remove = Gtk.ToolButton(stock_id=Gtk.STOCK_REMOVE)
        self.button_edit = Gtk.ToolButton(stock_id=Gtk.STOCK_EDIT)
        button_preferences = Gtk.ToolButton(stock_id=Gtk.STOCK_PREFERENCES)
        button_about = Gtk.ToolButton(stock_id=Gtk.STOCK_ABOUT)

        toolbar.add(self.button_play)
        toolbar.add(separator1)
        toolbar.add(button_dir)
        toolbar.add(button_add)
        toolbar.add(self.button_remove)
        toolbar.add(self.button_edit)
        toolbar.add(separator2)
        toolbar.add(button_preferences)
        toolbar.add(button_about)
        self.button_play.connect_object('clicked', self.on_play, self)
        button_dir.connect_object('clicked', self.on_add_dir, self)
        button_add.connect_object('clicked', self.on_add, self)
        self.button_remove.connect_object('clicked', self.on_remove, self)
        self.button_edit.connect_object('clicked', self.on_edit, self)
        button_about.connect_object('clicked', self.on_about, self)
        button_preferences.connect_object('clicked', self.on_preferences, self)

        self.button_play.set_sensitive(False)
        self.button_remove.set_sensitive(False)
        self.button_edit.set_sensitive(False)
        return toolbar

    def create_edit_dialog(self):
        '''This method constructs the dialog used to edit story entries.
    
        '''
        self.edit_dialog = Gtk.Dialog('Edit story', None,
                                      Gtk.DialogFlags.MODAL,
                                      (Gtk.STOCK_CANCEL,
                                       Gtk.ResponseType.REJECT,
                                       Gtk.STOCK_OK,
                                       Gtk.ResponseType.ACCEPT))
        self.edit_dialog.set_default_size(320, 480)
        fields = ['Title', 'Author', 'Group', 'Headline', 'First published',
                  'Language', 'Genre', 'Series', 'Series number',
                  'Forgiveness', 'Rating', 'Description', 'Format', 'IFID']
        dialog_table = Gtk.Table(len(fields), 3)
        for n, field in enumerate(fields):
            field_label = Gtk.Label(label='{0}: '.format(field))
            field_label.set_alignment(1, 0.5)
            if field == 'Description':
                field_buffer = Gtk.TextBuffer()
                field_view = Gtk.TextView(buffer=field_buffer)
                field_view.set_wrap_mode(Gtk.WrapMode.WORD)
                scroll = Gtk.ScrolledWindow()
                scroll.set_policy(Gtk.PolicyType.AUTOMATIC,
                                  Gtk.PolicyType.AUTOMATIC)
                scroll.add(field_view)
                dialog_table.attach(field_label, 0, 1, n, n + 1,
                                    xoptions=Gtk.AttachOptions.FILL)
                dialog_table.attach(scroll, 1, 2, n, n + 1,
                                    yoptions=Gtk.AttachOptions.FILL)
            elif field == 'Rating':
                combo_model = Gtk.ListStore(GdkPixbuf.Pixbuf, str)
                field_combo = Gtk.ComboBox(model=combo_model)
                combo_renderer = Gtk.CellRendererPixbuf()
                combo_renderer.set_padding(2, 5)
                combo_renderer.set_alignment(0, 0)
                field_combo.pack_start(combo_renderer, False)
                field_combo.add_attribute(combo_renderer, 'pixbuf', 0)
                field_combo.set_id_column(1)
                for rating in ['0', '0.5', '1', '1.5', '2', '2.5', '3',
                               '3.5', '4', '4.5', '5']:
                    pixbuf = self.library.get_rating_pixbuf(rating)
                    combo_model.append((pixbuf, rating))
                field_combo.set_active_id("0")
                dialog_table.attach(field_label, 0, 1, n, n + 1,
                                    xoptions=Gtk.AttachOptions.FILL)
                dialog_table.attach(field_combo, 1, 2, n, n + 1,
                                    #xoptions=Gtk.AttachOptions.EXPAND,
                                    yoptions=Gtk.AttachOptions.EXPAND)
            else:
                field_entry = Gtk.Entry()
                dialog_table.attach(field_label, 0, 1, n, n + 1,
                                    xoptions=Gtk.AttachOptions.FILL)
                dialog_table.attach(field_entry, 1, 2, n, n + 1)
        find_image = Gtk.Image.new_from_stock(Gtk.STOCK_FIND,
                                              Gtk.IconSize.MENU)
        ifid_search_button = Gtk.Button()
        ifid_search_button.set_image(find_image)
        ifid_search_button.connect('clicked', self.on_ifid_search)
        dialog_table.attach(ifid_search_button, 2, 3, 13, 14,
                            xoptions=Gtk.AttachOptions.SHRINK,
                            yoptions=Gtk.AttachOptions.SHRINK)
        vbox = self.edit_dialog.get_child()
        vbox.pack_start(dialog_table, True, True, 0)
        play_image = Gtk.Image.new_from_stock(Gtk.STOCK_MEDIA_PLAY,
                                              Gtk.IconSize.BUTTON)
        play_button = Gtk.Button()
        play_button.set_image(play_image)
        play_button.connect('clicked', self.on_edit_play_clicked)
        vbox.pack_start(play_button, False, False, 0)
        vbox.show_all()

    def load_library(self, library_filename):
        '''This method unpickles data file and creates a thread to 
        load the data into the library.

        '''
        try:
            with open(library_filename, 'rb') as library_handle:
                data = cPickle.load(library_handle)
        except:
            basename = os.path.splitext(library_filename)
            backup = '.'.join([basename[0], 'bak'])
            if os.path.exists(backup):
                backup_date = time.ctime(os.path.getmtime(library_filename))
                d = Gtk.MessageDialog(self, Gtk.DialogFlags.MODAL,
                                      Gtk.MessageType.QUESTION,
                                      Gtk.ButtonsType.YES_NO)
                d_msg1 = 'Oh no! Your library seems to have been corrupted.'
                d_msg2 = 'Would you like to restore a backup dated {0}?'
                d.set_markup('\n'.join(
                    [d_msg1, d_msg2.format(backup_date)]))
                restore_response = d.run()
                d.destroy()
                if restore_response == Gtk.ResponseType.YES:
                    shutil.copyfile(backup, library_filename)
                    try:
                        with open(library_filename, 'rb') as library_handle:
                            data = cPickle.load(library_handle)
                    except:
                        d = Gtk.MessageDialog(self, Gtk.DialogFlags.MODAL,
                                      Gtk.MessageType.WARNING,
                                      Gtk.ButtonsType.OK)
                        d_msg = 'Library recovery failed.'
                        d.set_markup(d_msg)
                        d.run()
                        d.destroy()
                        return False
                else:
                    return False
            else:
                d = Gtk.MessageDialog(self, Gtk.DialogFlags.MODAL,
                              Gtk.MessageType.WARNING,
                              Gtk.ButtonsType.OK)
                d_msg = 'Oh no! Your library is corrupted but no backup exists!'
                d.set_markup(d_msg)
                d.run()
                d.destroy()
                return False
        if len(data) == 0:
            return False
        if data.get('version') != GROTESQUE_VERSION:
            # If the save format changes in future versions, handle updating
            # the user's saved library here.
            #
            pass
        load_dialog = ProgressDialog("Loading library...", None, False)
        loading = libraryloadthread.load_library(self.library, load_dialog,
                                                 data)
        GObject.idle_add(loading.next)
        load_result = load_dialog.run()
        load_dialog.destroy()
        return True

    def save_library(self):
        '''This method opens a file handle to which to save the contents of the
        library.

        '''
        filename = self.settings.get_library_filename()
        if os.path.exists(filename):
            basename = os.path.splitext(filename)
            backup = '.'.join([basename[0], 'bak'])
            shutil.copyfile(filename, backup)
        with open(self.settings.get_library_filename(), 'wb') as lib_handle:
            try:
                self.library.save(lib_handle, GROTESQUE_VERSION)
            finally:
                lib_handle.close()

    def launch_story(self, story):
        '''This method handles launching story files.
    
        '''
        error_msg = None
        # First, check if there's an interpreter set for this kind of file.
        try:
            launcher = self.settings.get_launcher(story.format)
        except:
            error_msg1 = 'Failed to launch "{0}":'.format(story.title)
            error_msg2 = 'No interpreter configured for format {0}'.format(
                story.format)
            error_msg = '\n\n'.join([error_msg1, error_msg2])
        # Next, try to launch the interpreter with the file.
        else:
            try:
                Popen([launcher, story.story_file])
            except Exception, e:
                error_msg = 'Failed to launch "{0}":\n\n{1}'.format(
                    story.title, e.message)
        # If any errors arose, create a dialog to display them to the user.
        if error_msg != None:
            d = Gtk.MessageDialog(self, Gtk.DialogFlags.MODAL,
                                  Gtk.MessageType.WARNING, Gtk.ButtonsType.OK)
            d.set_markup(error_msg)
            d.run()
            d.destroy()

    def play_selection(self):
        '''This method grabs the currently selected story and sends it to be
        launched.
    
        '''
        selected_story = self.library_paned.get_selected_stories()[0]
        self.library.mark_story_played(selected_story[1])
        self.launch_story(selected_story[0])

    def remove_selection(self):
        '''This method grabs the currently selected stories and creates a
        thread to handle removing them from the library.
    
        '''
        row_iters = self.library_paned.get_selected_rows()
        if len(row_iters) > 0:
            remove_dialog = ProgressDialog("Removing...", None, False)
            #remove_thread = StoryRemoveThread(self.library, row_iters,
            #                                  remove_dialog)
            #remove_thread.start()
            removing = storyremovethread.remove_stories(self.library,
                                                        row_iters,
                                                        remove_dialog)
            GObject.idle_add(removing.next)
            remove_result = remove_dialog.run()
            #remove_thread.stop()
            remove_dialog.destroy()
        self.save_library()
        self.info_paned.clear()

    def do_import_dialog(self, filepaths):
        '''This method launches a thread which handles the actual importing of
        story files. If any files failed to be imported properly, the user has
        the option of manually editing their metadata.
        
        '''
        import_dialog = ProgressDialog("Importing...", self)
        import_thread = StoryImportThread(filepaths, self.settings,
                                          self.library, import_dialog)
        importing = import_thread.import_stories()
        GObject.idle_add(importing.next)
        import_response = import_dialog.run()
        import_dialog.destroy()
        if (import_response == Gtk.ResponseType.REJECT or
                import_response == Gtk.ResponseType.DELETE_EVENT):
            import_thread.stop()
        else:
            # If any files failed to import, create a dialog for editing them.
            if len(import_thread.fails) > 0:
                d = Gtk.MessageDialog(self, Gtk.DialogFlags.MODAL,
                                      Gtk.MessageType.QUESTION,
                                      Gtk.ButtonsType.YES_NO)
                d_msg1 = 'Metadata could not be retrieved for {0} files.'
                d_msg2 = 'Would you like to edit them now?'
                d.set_markup('\n'.join(
                    [d_msg1.format(len(import_thread.fails)), d_msg2]))
                edit_response = d.run()
                d.destroy()
                if edit_response == Gtk.ResponseType.YES:
                    vbox = self.edit_dialog.get_child()
                    for n, story in enumerate(import_thread.fails):
                        self.fill_edit_dialog(story)
                        self.edit_dialog.set_title(
                            "Edit story ({0}/{1})".format(
                                n + 1, len(import_thread.fails)))
                        response = self.edit_dialog.run()
                        if response == Gtk.ResponseType.ACCEPT:
                            table = vbox.get_children()[0]
                            items = table.get_children()
                            items.reverse()
                            story.title = items[1].get_text()
                            story.author = items[3].get_text()
                            story.group = items[5].get_text()
                            story.headline = items[7].get_text()
                            story.firstpublished = items[9].get_text()
                            story.language = items[11].get_text()
                            story.genre = items[13].get_text()
                            story.series = items[15].get_text()
                            story.seriesnumber = items[17].get_text()
                            story.forgiveness = items[19].get_text()
                            story.rating = items[21].get_active_id()
                            descr_view = items[23].get_children()[0]
                            descr_buffer = descr_view.get_buffer()
                            descr_start_iter = descr_buffer.get_start_iter()
                            descr_end_iter = descr_buffer.get_end_iter()
                            story.description = descr_buffer.get_text(
                                descr_start_iter, descr_end_iter, True)
                            story.format = items[25].get_text()
                            story.ifid = items[27].get_text()
                            self.library.add_entry(story)
                    self.edit_dialog.hide()
            # If any errors were encountered during import, display them here.
            if len(import_thread.msg) > 0:
                d = Gtk.MessageDialog(self, Gtk.DialogFlags.MODAL,
                                      Gtk.MessageType.WARNING,
                                      Gtk.ButtonsType.OK)
                d_msg = 'Errors encountered while adding games:\n\n{0}'.format(
                    import_thread.msg)
                d.set_markup(d_msg)
                d.run()
                d.destroy()
            self.save_library()

    def fill_edit_dialog(self, story):
        '''This method fills the edit dialog with the metadata currently stored
        in a story object.
    
        '''
        vbox = self.edit_dialog.get_child()
        table = vbox.get_children()[0]
        items = table.get_children()
        items.reverse()
        if story.title:
            items[1].set_text(story.title)
        else:
            items[1].set_text('Unknown')
        if story.author:
            items[3].set_text(story.author)
        else:
            items[3].set_text('Unknown')
        if story.group:
            items[5].set_text(story.group)
        else:
            items[5].set_text('None')
        if story.headline:
            items[7].set_text(story.headline)
        else:
            items[7].set_text('')
        if story.firstpublished:
            items[9].set_text(story.firstpublished)
        else:
            items[9].set_text('Unknown')
        if story.language:
            items[11].set_text(story.language)
        else:
            items[11].set_text('')
        if story.genre:
            items[13].set_text(story.genre)
        else:
            items[13].set_text('Unknown')
        if story.series:
            items[15].set_text(story.series)
        else:
            items[15].set_text('None')
        if story.seriesnumber:
            items[17].set_text(story.seriesnumber)
        else:
            items[17].set_text('')
        if story.forgiveness:
            items[19].set_text(story.forgiveness)
        else:
            items[19].set_text('Unknown')
        if story.starrating:
            items[21].set_active_id(story.starrating)
        else:
            items[21].set_active_id('0')
        description_view = items[23].get_children()[0]
        description_buffer = description_view.get_buffer()
        if story.description:
            description_buffer.set_text(story.description)
        else:
            description_buffer.set_text('')
        if story.format:
            items[25].set_text(story.format)
        else:
            items[25].set_text("Unknown")
        if story.ifid:
            items[27].set_text(story.ifid)
        else:
            items[27].set_text('')
        self.edit_story = story

    def on_vpaned_notify(self, vpaned, gparamspec):
        '''This method catches notify signals from the vpaned widget which
        separates the library view from the info view.  In particular, it is
        used to determine if the handle position has changed.  If it has
        changed, then the handle of the info paned should also be moved so that
        the coverart area is always square.
        
        '''
        if gparamspec.name == 'position':
            if self.init_complete and not self.vpaned_block:
                self.info_paned_block = True
                dimensions = self.get_size()
                self.info_paned.set_position(dimensions[1] -
                                             self.vpaned.get_position())
                self.info_paned.refresh_coverart()
                self.info_paned_block = False

    def on_infopaned_notify(self, hpaned, gparamspec):
        '''This method catches notify signals from the vpaned widget which
        separates the library view from the info view.  If the info paned
        handle is moved, then the vpaned handle should be moved as well to keep
        the cover art area square.

        '''
        if gparamspec.name == 'position':
            if self.init_complete and not self.info_paned_block:
                self.vpaned_block = True
                dimensions = self.get_size()
                self.vpaned.set_position(dimensions[1] -
                                         self.info_paned.get_position())
                self.info_paned.refresh_coverart()
                self.vpaned_block = False

    def on_key_pressed(self, widget, event):
        '''This method handles keypresses on the main window.
    
        '''
        if Gdk.keyval_name(event.keyval) == "Delete":
            self.remove_selection()
        elif Gdk.keyval_name(event.keyval) == "Return":
            self.play_selection()

    def on_remove(self, widget):
        '''This method handles the remove toolbar button being clicked.
    
        '''
        self.remove_selection()

    def on_play(self, widget):
        '''This method handles the play toolbar button being clicked.

        '''
        self.play_selection()

    def on_add(self, widget):
        '''This method handles the add toolbar button ('+' icon, single file)
        being clicked.
        
        '''
        exts = []
        all_exts = self.settings.get_all_exts()
        for if_format, format_exts in all_exts:
            exts.extend(format_exts.split(','))
        exts_set = set(exts)
        # Create a file filter to only show valid story files.
        story_filter = Gtk.FileFilter()
        story_filter.set_name("Story files")
        for ext in exts_set:
            story_filter.add_pattern('*.{0}'.format(ext))
        all_filter = Gtk.FileFilter()
        # Create a filter to show all files.
        all_filter.set_name("All files")
        all_filter.add_pattern("*")
        # Create the file chooser dialog and add the filters.
        file_chooser = Gtk.FileChooserDialog(
            "Select game files to add", self, Gtk.FileChooserAction.OPEN,
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OPEN,
             Gtk.ResponseType.ACCEPT))
        file_chooser.set_select_multiple(True)
        file_chooser.set_current_folder(self.settings.get_game_dir())
        file_chooser.add_filter(story_filter)
        file_chooser.add_filter(all_filter)
        # Run the dialog
        response = file_chooser.run()
        file_chooser.hide()
        if response == Gtk.ResponseType.ACCEPT:
            # Save this directory as the last-accessed one
            self.settings.set_game_dir(file_chooser.get_current_folder())
            self.settings.save()
            # Get the selected files.
            filepaths = file_chooser.get_filenames()
            if len(filepaths) > 0:
                # Launch a dialog to handle the import.
                self.do_import_dialog(filepaths)
        file_chooser.destroy()

    def on_add_dir(self, widget):
        '''This method handles the add directory toolbar (folder icon, 
        recursive import) being clicked.
        
        '''
        # Create a list of extensions to handle.
        exts = []
        all_exts = self.settings.get_all_exts()
        for if_format, format_exts in all_exts:
            exts.extend(format_exts.split(','))
        exts_set = set(exts)
        # Create and lanch the file chooser.
        file_chooser = Gtk.FileChooserDialog(
            "Select a directory to add", self,
            Gtk.FileChooserAction.SELECT_FOLDER, (Gtk.STOCK_CANCEL,
            Gtk.ResponseType.CANCEL, Gtk.STOCK_OPEN, Gtk.ResponseType.ACCEPT))
        file_chooser.set_select_multiple(False)
        file_chooser.set_current_folder(self.settings.get_game_dir())
        response = file_chooser.run()
        file_chooser.hide()
        if response == Gtk.ResponseType.ACCEPT:
            filepaths = []
            self.settings.set_game_dir(file_chooser.get_current_folder())
            self.settings.save()
            directory = file_chooser.get_filename()
            for dirpath, dirnames, filenames in os.walk(directory):
                # For all the files found, if they have the correct extension,
                # add them for import.
                if len(filenames) > 0:
                    for filename in filenames:
                        extension = filename.rpartition('.')[2]
                        if extension in exts_set:
                            filepath = os.path.join(dirpath, filename)
                            filepaths.append(filepath)
            if len(filepaths) > 0:
                # Assuming files were found, run the import dialog.
                self.do_import_dialog(filepaths)
        file_chooser.destroy()

    def on_edit(self, widget):
        '''This method handles the edit button being clicked.
        
        '''
        selected_stories = self.library_paned.get_selected_stories()
        for n, selected_story in enumerate(selected_stories):
            story, row_iter = selected_story
            self.fill_edit_dialog(story)
            self.edit_dialog.set_title('Edit story ({0}/{1})'.format(
                n + 1, len(selected_stories)))
            response = self.edit_dialog.run()
            if response == Gtk.ResponseType.ACCEPT:
                vbox = self.edit_dialog.get_child()
                table = vbox.get_children()[0]
                items = table.get_children()
                items.reverse()
                biblio = {}
                for n in range(0, len(items[:-1]), 2):
                    key = items[n].get_text().split(':')[0]
                    if key == 'Description':
                        scroll = items[n + 1]
                        field_view = scroll.get_children()[0]
                        field_buffer = field_view.get_buffer()
                        start_iter = field_buffer.get_start_iter()
                        end_iter = field_buffer.get_end_iter()
                        value = field_buffer.get_text(start_iter, end_iter,
                                                      True)
                    elif key == 'Rating':
                        value = items[n + 1].get_active_id()
                    else:
                        value = items[n + 1].get_text()
                    if value == "":
                        value = None
                    biblio[key] = value
                self.library.edit_entry(row_iter, biblio)
        self.edit_dialog.hide()
        self.save_library()
        self.info_paned.show_story(selected_stories[0][0])

    def on_preferences(self, widget):
        '''This method handles clicks on the preferences button.
        
        '''
        settings_dialog = SettingsDialog(self.settings, self)
        settings_dialog.run()
        settings_dialog.destroy()

    def on_about(self, widget):
        '''This method handles clicks on the about button.
        
        '''
        about = Gtk.AboutDialog()
        about.set_program_name('Grotesque')
        about.set_version(' '.join(['Version', GROTESQUE_VERSION]))
        comment = 'Organize and explore your Interactive Fiction library.'
        about.set_comments(comment)
        copyright_per = 'Copyright © 2009, 2010 Per Liedman'
        copyright_brandon = 'Copyright © 2011 Brandon Invergo'
        about.set_copyright('\n'.join([copyright_per, copyright_brandon]))
        about.set_license_type(Gtk.License.GPL_3_0)
        about.set_website('http://grotesque.sourceforge.net')
        about.set_website_label('Grotesque Website')
        about.set_logo(self.get_icon())
        about.set_authors(['Brandon Invergo', '\n', 'Past Developers:',
                           'Per Liedman'])
        about.show_all()
        about.run()
        about.destroy()

    def on_edit_play_clicked(self, button):
        '''The edit dialog has a play button on it, so the user can launch a
        file while editing it to help figure out some information. This method
        handles clicks on it.
        
        '''
        self.launch_story(self.edit_story)

    def on_ifid_search(self, button):
        '''The edit dialog has a search button on it. This method handles
        clicks on that button. It grabs the currently entered IFID and searches
        IFDB with it.
    
        '''
        vbox = self.edit_dialog.get_child()
        table = vbox.get_children()[0]
        items = table.get_children()
        items.reverse()
        ifid_entry = items[-2]
        ifid = ifid_entry.get_text().strip()
        if ifid == "":
            return False
        self.edit_story.ifid = ifid
        try:
            xmlfile = ifdb.fill_meta_from_ifdb(self.edit_story)
        except:
            #TODO do something more informative here
            return False
        coverart_dir = self.settings.coverart_dir
        ifdb.fetch_cover_art(self.edit_story, coverart_dir, xmlfile)
        self.fill_edit_dialog(self.edit_story)
        return True

    def on_library_selection_changed(self, library_paned):
        '''This method handles events in which the library selection changed.
        In particulary, this affects displaying info in the info paned and the
        active state of toolbar buttons.
        
        '''
        row_iters = self.library_paned.get_selected_rows()
        if len(row_iters) == 0:
            return
        if len(row_iters) == 1:
            story = self.library.get_story(row_iters[0])
            self.info_paned.show_story(story)
        elif len(row_iters) > 1:
            self.info_paned.clear()
        rows_selected = len(row_iters)
        self.button_play.set_sensitive(rows_selected == 1)
        self.button_edit.set_sensitive(rows_selected > 0)
        self.button_remove.set_sensitive(rows_selected > 0)

    def on_library_doubleclick(self, library_paned):
        self.play_selection()

    def on_close(self, widget, event, data=None):
        '''This method handles the window being closed. In particular, it
        handles saving settings and saving the library before exiting the main
        GTK loop.

        '''
        size = self.get_size()
        pos = self.vpaned.get_position()
        library_pos = self.library_paned.get_position()
        filter_type = self.library_paned.filter_combo.get_active()
        vpaned_percentage = float(pos) / float(size[1])
        library_percentage = float(library_pos) / float(size[0])
        self.settings.set_window_size(size)
        self.settings.set_vpaned_percentage(vpaned_percentage)
        self.settings.set_filter_percentage(library_percentage)
        self.settings.set_filter_type(filter_type)
        for col in self.library_paned.library_view.get_columns():
            width = col.get_width()
            title = col.get_title()
            self.settings.set_column_width(title, width)
            visible = col.get_visible()
            self.settings.set_column_visible(title, visible)
        self.settings.save()
        self.save_library()
        Gtk.main_quit()
        return False


def main():
    # If this is the first time running the program, the settings file won't
    # exist, so run teh settings assistant.
    settings = Settings()
    if not settings.load():
        from settingsassistant import SettingsAssistant
        setup_assistant = SettingsAssistant(settings, None, Gtk.main_quit)
        Gtk.main()
        try:
            os.makedirs(settings.appdir)
        except EnvironmentError, e:
            if e.errno != errno.EEXIST:
                raise
        settings.save()
    library_filename = settings.get_library_filename()
    main_wnd = MainWindow(library_filename, settings)
    Gtk.main()


if __name__ == "__main__":
    main()
