# -*- coding: utf-8 -*-
#
#       ifdb.py
#       
#       Copyright © 2008, 2009, 2010 Per Liedman <per@liedman.net>
#       Copyright © 2011 Brandon Invergo <b.invergo@gmail.com>
#       
#       This file is part of Grotesque.
#
#       Grotesque is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       Grotesque is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with Grotesque.  If not, see <http://www.gnu.org/licenses/>.


import xml.sax
from urllib2 import urlopen
from urllib import urlretrieve
import os.path
import os
#for python3: from urllib.request import urlretrieve

from storydata import StoryData


class IfdbHandler(xml.sax.ContentHandler):
    '''This class implements an XML handler for the data retrieved from the
    IFDB.
    
    '''
    def __init__(self):
        self.story = StoryData()
        self.current_tag = None
        self.char_buf = []
        self.coverart_url = None
        self.coverart_tag = False

    def characters(self, data):
        self.char_buf.append(data)

    def startElement(self, tag, attrs):
        if tag == 'coverart':
            self.coverart_tag = True

    def endElement(self, tag):
        if tag == 'br':
            return
        text = "".join(self.char_buf)
        text = text.replace('&amp;', '&')
        text = text.replace('&quot;', '"')
        if tag == 'title':
            self.story.title = text.encode('UTF-8')
        elif tag == 'author':
            self.story.author = text.encode('UTF-8')
        elif tag == 'language':
            self.story.language = text.encode('UTF-8')
        elif tag == 'firstpublished':
            self.story.firstpublished = text.encode('UTF-8')
        elif tag == 'genre':
            self.story.genre = text.encode('UTF-8')
        elif tag == 'series':
            self.story.series = text.encode('UTF-8')
        elif tag == 'seriesnumber':
            self.story.seriesnumber = text.encode('UTF-8')
        elif tag == 'forgiveness':
            self.story.forgiveness = text.encode('UTF-8')
        elif tag == 'headline':
            self.story.headline = text.encode('UTF-8')
        elif tag == 'group':
            self.story.group = text.encode('UTF-8')
        elif tag == 'description':
            self.story.description = text.encode('UTF-8')
        elif tag == 'starRating':
            self.story.starrating = text.encode('UTF-8')
        elif tag == 'coverart':
            self.coverart_tag = False
        elif tag == 'url':
            if self.coverart_tag:
                self.coverart_url = text
        self.char_buf = []


def fill_meta_from_ifdb(story):
    '''This method retrieves metadata from IFDB for a given story and fills the
    story object with it
    
    '''
    handler = IfdbHandler()
    # Access the IFDB public API
    url = ''.join(['http://ifdb.tads.org/viewgame?ifiction&ifid=', story.ifid])
    xmlfile = urlopen(url)
    xml.sax.parse(xmlfile, handler)
    story.title = handler.story.title
    story.author = handler.story.author
    story.language = handler.story.language
    story.firstpublished = handler.story.firstpublished
    story.genre = handler.story.genre
    story.series = handler.story.series
    story.seriesnumber = handler.story.seriesnumber
    story.forgiveness = handler.story.forgiveness
    story.group = handler.story.group
    story.description = handler.story.description
    story.starrating = handler.story.starrating


def fetch_cover_art(story, coverart_dir, xmlfile=None):
    '''This method fetches coverart from the IFDB for a given story. If an XML
    file for the story has previously been retrieved, it can be passed here to
    avoid repeat requests to the serven.
    
    '''
    if not story.ifid:
        return None
    handler = IfdbHandler()
    if xmlfile is None:
        url = ''.join(['http://ifdb.tads.org/viewgame?ifiction&ifid=',
                      story.ifid])
        xmlfile = urlopen(url)
    xml.sax.parse(xmlfile, handler)
    coverart_url = handler.coverart_url
    if coverart_url:
        if not os.path.exists(coverart_dir):
            os.mkdir(coverart_dir)
        coverart_file = os.path.join(coverart_dir, ''.join([story.ifid, '.jpg']))
        if not os.path.exists(coverart_file):
            urlretrieve(coverart_url, coverart_file)
    return xmlfile


def fetch_rating(story, xmlfile=None):
    '''This method fetches ratings from the IFDB for a given story. If an XML
    file for the story has previously been retrieved, it can be passed here to
    avoid repeat requests to the serven.
    
    '''
    if not story.ifid:
        return None
    handler = IfdbHandler()
    if xmlfile is None:
        url = ''.join(['http://ifdb.tads.org/viewgame?ifiction&ifid=',
                      story.ifid])
        xmlfile = urlopen(url)
    xml.sax.parse(xmlfile, handler)
    story.starrating = handler.story.starrating
    return xmlfile
