# -*- coding: utf-8 -*-
#
#       main.py
#       
#       Copyright © 2009, 2010 Per Liedman <per@liedman.net>
#       Copyright © 2011 Brandon Invergo <b.invergo@gmail.com>
#       
#       This file is part of Grotesque.
#
#       Grotesque is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       Grotesque is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with Grotesque.  If not, see <http://www.gnu.org/licenses/>.


import struct
import re
import md5
from binascii import hexlify

from storydata import StoryData
from ifformaterror import IFFormatError


T2_SIGNATURE = 'TADS2 bin\012\015\032'
T3_SIGNATURE = 'T3-image\015\012\032'
HTML_RES_ID = 'HTMLRES'
EOF_RES_ID = '$EOF'


attribs = [('name', 'title'), ('ifid', 'ifid'),
           ('byline', 'author', re.compile('[Bb]y (.+)')),
           ('headline', 'headline'), ('desc', 'description'),
           ('firstpublished', 'first_published'), ('language', 'language'),
           ('series', 'series'), ('seriesnumber', 'seriesnumber'),
           ('genre', 'genre'), ('forgiveness', 'foregivness')]


def get_gameinfo(file_buffer):
    rsc = find_resource(file_buffer, 'GameInfo.txt')
    if rsc is not None:
        rsc = rsc.decode('utf-8')
        gameinfo = {}
        for line in rsc.split('\n'):
            line = line.strip()
            if not line.startswith('#'):
                parts = line.split(':', 1)
                if len(parts) == 2:
                    key = parts[0].strip().lower()
                    value = parts[1].strip().replace('\\n', '\n')
                    gameinfo[key] = value
        return gameinfo
    else:
        return None


def find_resource(file_buffer, name):
    if is_tads2(file_buffer):
        return t2_find_resource(file_buffer, name)
    elif is_tads3(file_buffer):
        return t3_find_resource(file_buffer, name)
    else:
        raise IFFormatError('file_buffer is neither TADS2 or TADS3.')


def t2_find_resource(file_buffer, name):
    name = name.lower()

    # Skip past TADS2 header (13 bytes signature, 7 bytes version, 2 bytes
    # flags, 26 bytes timestamp
    p = 13 + 7 + 2 + 26
    while p < len(file_buffer):
        type_length = read_byte(file_buffer, p)
        next_section = read_int(file_buffer, p + 1 + type_length)
        type = file_buffer[p + 1:p + 1 + type_length]
        if type == HTML_RES_ID:
            # Skip past the section block header
            p = p + 12
            number_entries = read_int(file_buffer, p)
            # Skip the first index entry
            p = p + 8
            # Scan index entries
            found_offset = None
            found_length = None
            for x in range(1, number_entries):
                rsc_offset = read_int(file_buffer, p)
                rsc_size = read_int(file_buffer, p + 4)
                name_len = read_short(file_buffer, p + 8)
                p = p + 10
                rsc_name = file_buffer[p:p + name_len]
                if name_len == len(name) and rsc_name.lower() == name:
                    found_offset = rsc_offset
                    found_length = rsc_size
                p = p + name_len
            if found_offset is not None:
                return file_buffer[p + found_offset:p + found_offset +
                                   found_length]
        elif type == EOF_RES_ID:
            return None
        p = next_section
    return None


def t3_find_resource(file_buffer, name):
    name = name.lower()

    # Skip past TADS3 header (11 bytes signature, 2 bytes version, 32 bytes
    # reserved, 24 bytes timestamp
    p = 11 + 2 + 32 + 24
    while p < len(file_buffer):
        type = file_buffer[p:p + 4]
        block_size = read_int(file_buffer, p + 4)
        if type == 'MRES':
            # Skip past the section block header
            p = p + 10
            block_start = p
            number_entries = read_short(file_buffer, p)
            p = p + 2
            # Scan index entries
            for x in range(1, number_entries):
                rsc_offset = read_int(file_buffer, p)
                rsc_size = read_int(file_buffer, p + 4)
                name_len = read_byte(file_buffer, p + 8)
                rsc_name = ''
                for xored_char in file_buffer[p + 9:p + 9 + name_len]:
                    rsc_name = rsc_name + chr(ord(xored_char) ^ 0xFF)

                if name_len == len(name) and rsc_name.lower() == name:
                    return file_buffer[block_start + rsc_offset:\
                                      block_start + rsc_offset + rsc_size]
                p = p + 9 + name_len
            p = block_start + block_size
        elif type == 'EOF':
            return None
        else:
            p = p + 10 + block_size
    return None


def get_extensions():
    return ['gam', 't3']


def is_tads2(file_buffer):
    return file_buffer.startswith(T2_SIGNATURE)


def is_tads3(file_buffer):
    return file_buffer.startswith(T3_SIGNATURE)


def is_a(file_buffer):
    return is_tads2(file_buffer) or is_tads3(file_buffer)


def get_information(file_buffer):
    s = StoryData()
    if is_tads2(file_buffer):
        s.format = 'tads2'
    elif is_tads3(file_buffer):
        s.format = 'tads3'
    else:
        raise IFFormatError('file_buffer is neither TADS2 or TADS3.')
    gameinfo = get_gameinfo(file_buffer)
    if gameinfo is not None:
        for attrib in attribs:
            value = gameinfo.get(attrib[0])
            if value is not None:
                if len(attrib) > 2:
                    match = attrib[2].match(value)
                    if match is not None:
                        value = match.group(1)
                setattr(s, attrib[1], value)
    if s.ifid is None:
        file_hash = hexlify(md5.new(file_buffer).digest()).upper()
        if is_tads2(file_buffer):
            s.ifid = ''.join(['TADS2-', file_hash])
        elif is_tads3(file_buffer):
            s.ifid = ''.join(['TADS3-', file_hash])
        else:
            raise IFFormatError('file_buffer is neither TADS2 or TADS3.')
    return s


def read_int(file_buffer, p):
    return struct.unpack_from('<I', file_buffer, p)[0]


def read_short(file_buffer, p):
    return struct.unpack_from('<H', file_buffer, p)[0]


def read_byte(file_buffer, p):
    return struct.unpack_from('<B', file_buffer, p)[0]


def matches(file_buffer, offset, string):
    return file_buffer[offset:offset + len(string)] == string
